#!/usr/bin/env python3

import sys
import os
import argparse
import os.path
import json
import pandas as pd
import numpy
import re
import pandas
def eprint(*args, **kwargs): print(*args, file=sys.stderr, **kwargs)


####################################################################################################
#### SDRF handler class for reading, writing, and merging SDRFs
class SDRFHandler:


    ####################################################################################################
    #### Constructor
    def __init__(self, filepath=None, verbose=None):

        self.filepath = filepath
        self.filename = self.extract_filename(filepath)
        self.filetype = { 'extension': None, 'type': None, 'delimiter': None }

        self.sdrf_data = None
        self.column_properties = None
        self.n_rows = None
        self.n_samples = None
        self.n_assays = None
        self.n_files = None

        self.ui_url = None
        self.data_url = None
        self.problems = { 
            'errors': { 'count': 0, 'codes': {}, 'list': [] },
            'warnings': { 'count': 0, 'codes': {}, 'list': [] }
        }
 
        #### Check if a verbose was provided and if not, set to default
        if verbose is None:
            verbose = 0
        self.verbose = verbose



    ####################################################################################################
    #### Extract the file name portion from a full path
    def extract_filename(self, filepath):

        if filepath is None:
            return

        if not isinstance(filepath, str):
            eprint(f"ERROR: SDRFData.extract_filename: filepath must be a string")
            return

        match = re.match(r'.*/(.+)?$', filepath)
        if match:
            filename = match.group(1)

        else:
            filename = filepath

        self.filename = filename
        return filename



    ####################################################################################################
    #### Try to infer some things about the file based on the name
    def infer_filetype(self):

        filename = self.filename
        filetype = { 'extension': None, 'type': None, 'delimiter': None }

        if filename.casefold().endswith('tsv'.casefold()):
            filetype = { 'extension': filename[-3:], 'type': 'tsv', 'delimiter': "\t" }
        elif filename.casefold().endswith('txt'.casefold()):
            filetype = { 'extension': filename[-3:], 'type': 'tsv', 'delimiter': "\t" }
        elif filename.casefold().endswith('sdrf'.casefold()):
            filetype = { 'extension': filename[-4:], 'type': 'tsv', 'delimiter': "\t" }
        elif filename.casefold().endswith('csv'.casefold()):
            filetype = { 'extension': filename[-3:], 'type': 'csv', 'delimiter': "," }
            self.log_problem('ERROR', 'CSVNotValid', 0, 0, f"SDRF must be tab-separated, not comma-separated")
        elif filename.casefold().endswith('xls'.casefold()):
            filetype = { 'extension': filename[-3:], 'type': 'xls', 'delimiter': None }
            self.log_problem('ERROR', 'XLSNotValid', 0, 0, f"SDRF must be tab-separated, not Excel xls format")
        elif filename.casefold().endswith('xlsx'.casefold()):
            filetype = { 'extension': filename[-4:], 'type': 'xls', 'delimiter': None }
            self.log_problem('ERROR', 'XLSXNotValid', 0, 0, f"SDRF must be tab-separated, not Excel xlsx format")
        else:
            eprint(f"WARNING: SDRFData.infer_filetype: Unable to guess the formatting based on filename. Guessing TSV")
            filetype = { 'extension': '', 'type': 'tsv', 'delimiter': "\t" }

        self.filetype = filetype
        return filetype



    ####################################################################################################
    #### Read an SDRF file
    def read(self, filepath=None):

        if filepath is not None:
            self.filepath = filepath
            self.filename = self.extract_filename(filepath)
        else:
            filepath = self.filepath

        if filepath is None:
            eprint(f"ERROR: SDRFData.read: No available filepath to read")
            return

        filetype = self.infer_filetype()

        if filetype['delimiter'] is not None:
            with open(filepath, encoding='utf-8-sig', errors='ignore') as infile:
                first_line = True
                sdrf_data = { 'titles': None, 'rows': [] }
                for line in infile:
                    if len(line.strip()) < 2:
                        continue
                    if first_line:
                        sdrf_data['titles'] = line.strip().split(filetype['delimiter'])
                        first_line = False
                        continue
                    columns = line.strip().split(filetype['delimiter'])
                    #### Add empty values at the end if the row has fewer columns of data than the titles
                    while len(columns) < len(sdrf_data['titles']):
                        columns.append('')
                    sdrf_data['rows'].append(columns)

        elif filetype['type'] == 'xls':
            source_sdrf_data = pandas.read_excel(filepath)
            sdrf_data = { 'titles': list(source_sdrf_data), 'rows': [] }
            for index, row in source_sdrf_data.iterrows():
                rowdata = row.tolist()
                #### Deal with NaN values, which cause problems
                newrowdata = []
                icolumn = 1
                for value in rowdata:
                    if isinstance(value, float) and numpy.isnan(value):
                        value = 'NaN'
                        self.log_problem('ERROR', 'NaNValue', index + 1, icolumn, f"Value cannot be NaN")
                    newrowdata.append(value)
                    icolumn += 1
                sdrf_data['rows'].append(newrowdata)

        else:
            eprint(f"WARNING: Don't know how to read file '{filepath}'")
            return

        self.sdrf_data = sdrf_data

        self.check_and_fix_headers()
        self.compute_stats()



    ####################################################################################################
    #### Check and potentially fix the header values
    def check_and_fix_headers(self):

        sdrf_data = self.sdrf_data
        if sdrf_data is None:
            eprint(f"ERROR: SDRFData.check_and_fix_headers: No available data to process")
            return

        required_columns = {
            'source name': None,
            'assay name': None,
            'comment[fraction identifier]': None,
            'comment[label]': None,
            'comment[data file]': None,
        }

        recommended_columns = {
            'technology type': None,
            'characteristics[organism]': None,
            'characteristics[organism part]': None,
            'comment[technical replicate]': None,
            'comment[biological replicate]': None,
            'comment[file uri]': None,
            'comment[modification parameters]': None,
            'comment[cleavage agent details]': None,
            'comment[instrument]': None,
        }

        titles = sdrf_data['titles']
        new_titles = []
        icolumn = 0
        for title in titles:
            if title != title.lower():
                self.log_problem('WARNING', 'TitleNotLowerCase', 0, icolumn, f"Column header '{title}' SHOULD be lower case")
            new_titles.append(title.lower())
            if icolumn == 0 and title.lower() != 'source name':
                self.log_problem('ERROR', 'Column0NotSourceName', 0, icolumn, f"The first column title MUST be 'source name'")
            if title.strip() == '':
                self.log_problem('ERROR', 'EmptyColumnTitle', 0, icolumn, f"Column {icolumn+1} title is empty")
            icolumn += 1
        sdrf_data['titles'] = new_titles

        icolumn = 0
        for title in new_titles:
            if title in required_columns:
                required_columns[title] = icolumn
            if title in recommended_columns:
                recommended_columns[title] = icolumn
            icolumn += 1

        for title, icolumn in required_columns.items():
            if icolumn is None:
                self.log_problem('ERROR', 'MissingRequiredColumn', 0, 0, f"Column {title} is missing")
        for title, icolumn in recommended_columns.items():
            if icolumn is None:
                self.log_problem('WARNING', 'MissingRecommendedColumn', 0, 0, f"Column {title} is missing")



    ####################################################################################################
    #### Check and potentially fix the header values
    def compute_stats(self):

        sdrf_data = self.sdrf_data
        if sdrf_data is None:
            eprint(f"ERROR: SDRFData.check_and_fix_headers: No available data to process")
            return

        columns_of_interest = [ 'source name', 'assay name', 'comment[data file]', 'comment[file uri]' ]
        column_indexes = {}
        column_unique_values = {}

        for column in columns_of_interest:
            column_indexes[column] = None
            column_unique_values[column] = {}

        titles = sdrf_data['titles']
        column_properties = []
        icolumn = 0
        for title in titles:
            if title in column_indexes:
                column_indexes[title] = icolumn
            column_properties.append( { 'is_empty': True } )
            icolumn += 1

        for row in sdrf_data['rows']:
            for title in columns_of_interest:
                if column_indexes[title] is not None:
                    icolumn = column_indexes[title]
                    column_unique_values[title][row[icolumn]] = True

        #### Record which columns are completely empty (i.e. no values in the column)
        for row in sdrf_data['rows']:
            icolumn = 0
            for title in titles:
                if row[icolumn] is not None and row[icolumn] != '':
                    column_properties[icolumn]['is_empty'] = False
                icolumn += 1
        self.column_properties = column_properties

        self.n_rows = len(sdrf_data['rows'])
        if self.n_rows == 0:
            self.log_problem('ERROR', 'ZeroRows', 0, 0, f"SDRF file has no data rows")

        self.n_samples = len(column_unique_values['source name'])
        if self.n_samples == 0:
            self.log_problem('ERROR', 'ZeroSamples', 0, 0, f"SDRF file has no samples information in 'source name'")

        self.n_assays = len(column_unique_values['assay name'])
        if self.n_assays == 0:
            self.log_problem('ERROR', 'ZeroAssays', 0, 0, f"SDRF file has no assay information for 'assay name'")

        self.n_files = len(column_unique_values['comment[data file]'])
        if len(column_unique_values['comment[file uri]']) > self.n_files:
            self.n_files = len(column_unique_values['comment[file uri]'])
            self.log_problem('ERROR', 'MissingFileNames', 0, column_indexes['comment[data file]'], f"Data for file names is missing (no data for 'comment[data file]')")
        if self.n_files == 0:
            self.log_problem('ERROR', 'MissingFileData', 0, 0, f"Neither data file nor file uri is present")



    ####################################################################################################
    #### Merge the provided SDRF object into the current SDRF object
    def merge(self, merge_sdrf, tag_with_decisions=None, prefer_main_columns=None):

        merge_empty_columns = False
        preserve_main_source_name = True

        if self.sdrf_data is None:
            eprint(f"ERROR: SDRFData.merge: No available data to process")
            return

        if merge_sdrf is None:
            eprint(f"ERROR: SDRFData.merge: Provided sdrf to merge is null. Cannot continue")
            return

        if self.n_files != merge_sdrf.n_files:
            self.log_problem('ERROR', 'DifferingNFiles', 0, 0, f"Primary SDRF has {self.n_files} files and SDRF to merge has {merge_sdrf.n_files} files. Cannot merge such differing SDRFs")
            return

        if self.n_rows != merge_sdrf.n_rows:
            self.log_problem('ERROR', 'DifferingNRows', 0, 0, f"Primary SDRF has {self.n_rows} files and SDRF to merge has {merge_sdrf.n_rows} files. Cannot merge such differing SDRFs")
            return

        if prefer_main_columns is None or prefer_main_columns == '':
            prefer_main_columns = []
        else:
            prefer_main_columns = prefer_main_columns.split(',')

        #### Set up a data structure to capture information about how to merge
        merge_data = { 'match_data': {}, 'columns': [] }

        #### Define primary columns that could be used for merging
        columns_of_interest = [ 'source name', 'assay name', 'comment[data file]', 'comment[file uri]' ]
        for title in columns_of_interest:
            if title not in merge_data['match_data']:
                merge_data['match_data'][title] = { 'main_icolumn': None, 'main_values': {}, 'merge_icolumn': None, 'merge_values': {} }

        #### Scan for and record these merge reference columns
        icolumn = 0
        for title in self.sdrf_data['titles']:
            if title in merge_data['match_data']:
                merge_data['match_data'][title]['main_icolumn'] = icolumn
            icolumn += 1

        icolumn = 0
        for title in merge_sdrf.sdrf_data['titles']:
            if title in merge_data['match_data']:
                merge_data['match_data'][title]['merge_icolumn'] = icolumn
            icolumn += 1

        #### Collate the row-level values for the target SDRF
        for row in self.sdrf_data['rows']:
            for title in columns_of_interest:
                icolumn = merge_data['match_data'][title]['main_icolumn']
                #print(f"title={title} main_icolumn={merge_data['match_data'][title]['main_icolumn']}")
                if icolumn is not None:
                    value = row[icolumn]
                    if value not in merge_data['match_data'][title]['main_values']:
                        merge_data['match_data'][title]['main_values'][value] = 0
                    merge_data['match_data'][title]['main_values'][value] += 1
                merge_data['match_data'][title]['main_n_values'] = len(merge_data['match_data'][title]['main_values'])

        #### Collate the row-level values for the SDRF to merge in
        for row in merge_sdrf.sdrf_data['rows']:
            for title in columns_of_interest:
                icolumn = merge_data['match_data'][title]['merge_icolumn']
                #print(f"title={title} merge_icolumn={merge_data['match_data'][title]['merge_icolumn']}")
                if icolumn is not None:
                    value = row[icolumn]
                    if value not in merge_data['match_data'][title]['merge_values']:
                        merge_data['match_data'][title]['merge_values'][value] = 0
                    merge_data['match_data'][title]['merge_values'][value] += 1
                merge_data['match_data'][title]['merge_n_values'] = len(merge_data['match_data'][title]['merge_values'])

        #### Set up the columns from the main SDRF
        icolumn = 0
        for title in self.sdrf_data['titles']:
            iexisting_column = 0
            found_column = False
            for existing_column in merge_data['columns']:
                if title == existing_column['title']:
                    merge_data['columns'][iexisting_column]['main_icolumn'].append(icolumn)
                    found_column = True
                iexisting_column += 1

            if found_column is False:
                merge_data['columns'].append( { 'title': title, 'main_icolumn': [icolumn], 'merge_icolumn': [] } )

            icolumn += 1

        #### Mix in the columns from the merge SDRF
        icolumn = 0
        for title in merge_sdrf.sdrf_data['titles']:
            iexisting_column = 0
            found_column = False
            for existing_column in merge_data['columns']:
                if title == existing_column['title']:
                    if 'merge_icolumn' in merge_data['columns'][iexisting_column]:
                        merge_data['columns'][iexisting_column]['merge_icolumn'].append(icolumn)
                    else:
                        merge_data['columns'][iexisting_column]['merge_icolumn'] = [icolumn]
                    found_column = True
                iexisting_column += 1

            if found_column is False and merge_empty_columns is True:
                merge_data['columns'].append( { 'title': title, 'merge_icolumn': [icolumn], 'main_icolumn': [] } )

            icolumn += 1


        #### Consider if we have when it takes for a data file level merge
        title = 'comment[data file]'
        potential_suffixes = [ '.raw', '.RAW', '', '.mzML', '.mzml', '.d', '.D', '.Raw', '.wiff', '.WIFF', '.mzXML', '.mzxml' ]
        previously_found_suffix = None
        merge_data['match_data'][title]['merge_leftover_values'] = merge_data['match_data'][title]['merge_values'].copy()
        merge_data['match_data'][title]['main_leftover_values'] = merge_data['match_data'][title]['main_values'].copy()
        merge_data['match_data'][title]['adjusted_merge_values'] = []

        for value in merge_data['match_data'][title]['merge_values']:
            #eprint(f"--Matching {value}")
            if value in merge_data['match_data'][title]['main_values']:
                del(merge_data['match_data'][title]['merge_leftover_values'][value])
            else:
                #### Try to match without filename suffixes
                match = re.match(r'(.+)\.(.+)?', value)
                if match:
                    value_root = match.group(1)
                    suffix = match.group(2)
                    #eprint(f"      (decomposed to '{value_root}' and '{suffix}')")
                    if previously_found_suffix is not None:
                        #eprint(f"  - Trying {value_root + previously_found_suffix}")
                        if value_root + previously_found_suffix in merge_data['match_data'][title]['main_values']:
                            #eprint(f"    + Found {value_root + previously_found_suffix}")
                            del(merge_data['match_data'][title]['merge_leftover_values'][value])
                            value = value_root + previously_found_suffix
                    else:
                        for potential_suffix in potential_suffixes:
                            #eprint(f"  - Trying {value_root + potential_suffix}")
                            if value_root + potential_suffix in merge_data['match_data'][title]['main_values']:
                                #eprint(f"    + Found {value_root + potential_suffix}")
                                del(merge_data['match_data'][title]['merge_leftover_values'][value])
                                previously_found_suffix = potential_suffix
                                value = value_root + potential_suffix
                                break
            merge_data['match_data'][title]['adjusted_merge_values'].append(value)

        previously_found_suffix = None
        for value in merge_data['match_data'][title]['main_values']:
            if value in merge_data['match_data'][title]['merge_values']:
                del(merge_data['match_data'][title]['main_leftover_values'][value])
            else:
                #### Try to match without filename suffixes
                match = re.match(r'(.+)(\.+)?', value)
                if match:
                    value_root = match.group(1)
                    suffix = match.group(2)
                    if previously_found_suffix is not None:
                        if value_root + previously_found_suffix in merge_data['match_data'][title]['merge_values']:
                            del(merge_data['match_data'][title]['main_leftover_values'][value])
                    else:
                        for potential_suffix in potential_suffixes:
                            if value_root + potential_suffix in merge_data['match_data'][title]['merge_values']:
                                del(merge_data['match_data'][title]['main_leftover_values'][value])
                                previously_found_suffix = potential_suffix
                                break

        if len(merge_data['match_data'][title]['merge_leftover_values']) == len(merge_data['match_data'][title]['merge_values']):
            self.log_problem('ERROR', 'MismatchedDataFiles', 0, 0, f"There appears to be no correlation between the values in '{title}' between the two files. Cannot merge")
            return

        if len(merge_data['match_data'][title]['merge_values']) != merge_sdrf.n_rows:
            self.log_problem('ERROR', 'UnableToMerge', 0, 0, f"The current merge strategy requires that the merge file have one row per 'comment[data file]' in the merge SDRF. This is not the case, so cannot proceed for now.")
            return

        #### Index the merge rows
        irow = 0
        join_row_keys = {}
        join_title = 'comment[data file]'
        icolumn = merge_data['match_data'][join_title]['merge_icolumn']
        for row in merge_sdrf.sdrf_data['rows']:
            #### Rewrite the merge column value in the merge data to match the main value for the merge column
            row[icolumn] = merge_data['match_data'][title]['adjusted_merge_values'][irow]
            value = row[icolumn]
            #value = merge_data['match_data'][title]['adjusted_merge_values'][irow]
            #eprint(f"%%%%  {irow}: {value}")
            join_row_keys[value] = irow
            irow += 1

        #### Create the final column title list for the merged result
        merged_data = { 'titles': [], 'rows': [] }
        for merged_column in merge_data['columns']:
            title = merged_column['title']
            if len(merged_column['main_icolumn']) == 1:
                merged_data['titles'].append(title)
            elif len(merged_column['main_icolumn']) == 0 and len(merged_column['merge_icolumn']) == 1:
                merged_data['titles'].append(title)
            elif len(merged_column['main_icolumn']) > 1:
                for icolumn in merged_column['main_icolumn']:
                    merged_data['titles'].append(title)
            elif len(merged_column['merge_icolumn']) > 1:
                for icolumn in merged_column['merge_icolumn']:
                    merged_data['titles'].append(title)
            else:
                eprint(f"ERROR: Unable to find a matching pattern in the number of columns for column label '{title}'")
                exit()

        #### Loop over all the rows and merge in the information
        irow = 0
        for row in self.sdrf_data['rows']:
            icolumn = 0
            new_row = []

            join_value = row[ merge_data['match_data'][join_title]['main_icolumn'] ]
            join_irow = join_row_keys[join_value]
            join_row = merge_sdrf.sdrf_data['rows'][join_irow]

            for merged_column in merge_data['columns']:
                title = merged_column['title']
                store_value = True
                if len(merged_column['main_icolumn']) == 1 and len(merged_column['merge_icolumn']) == 0:
                    value = row[merged_column['main_icolumn'][0]]
                    if tag_with_decisions:
                        value += "_onlyMain"
                elif len(merged_column['main_icolumn']) == 0 and len(merged_column['merge_icolumn']) == 1:
                    value = join_row[merged_column['merge_icolumn'][0]]
                    if tag_with_decisions:
                        value += "_onlyMerge"
                elif len(merged_column['main_icolumn']) == 1 and len(merged_column['merge_icolumn']) == 1:
                    main_value = row[merged_column['main_icolumn'][0]]
                    merge_value = join_row[merged_column['merge_icolumn'][0]]
                    if main_value == merge_value:
                        value = main_value
                        if tag_with_decisions:
                            value += "_same"
                    elif main_value == '' and merge_value > '':
                        value = merge_value
                        if tag_with_decisions:
                            value += "_mergeNonempty"
                    elif main_value > '' and merge_value == '':
                        value = main_value
                        if tag_with_decisions:
                            value += "_mainNonempty"
                    elif main_value == '' and merge_value == '':
                        value = main_value
                    else:
                        if title in prefer_main_columns:
                            value = main_value
                            if tag_with_decisions:
                                value += "_UserRequestKeepMain"
                            self.log_problem('WARNING', 'UserRequestKeepMain', irow + 1, icolumn + 1, f"For column '{title}', values in main ({main_value}) and merge ({merge_value}) differ. By user request, keeps the main.")
                        elif title == 'source name' and preserve_main_source_name:
                            value = main_value
                            if tag_with_decisions:
                                value += "_SpecialRuleKeepMain"
                            self.log_problem('WARNING', 'SpecialRuleKeepMain', irow + 1, icolumn + 1, f"For column '{title}', values in main ({main_value}) and merge ({merge_value}) differ. A special rule always keeps the main.")
                        else:
                            value = merge_value
                            if tag_with_decisions:
                                value += "_preferMergeOverMain"
                            self.log_problem('WARNING', 'MergeValueOverridesMain', irow + 1, icolumn + 1, f"For column '{title}', values in main ({main_value}) and merge ({merge_value}) differ. Value from merge ({merge_value}) supercedes.")
                elif len(merged_column['main_icolumn']) > 1:
                    for imulti_column in merged_column['main_icolumn']:
                        value = row[merged_column['merge_icolumn'][0]]
                        if tag_with_decisions:
                            value += '_multi'
                        new_row.append(value)
                    store_value = False
                else:
                    eprint(f"Ack. need to figure out what to do with {title} and {merged_column['main_icolumn']}")
                    exit()
                if store_value:
                    new_row.append(value)
                icolumn += 1
            merged_data['rows'].append(new_row)
            irow += 1

        self.sdrf_data = merged_data
        self.compute_stats()

        #print(json.dumps(merge_data, indent=2, sort_keys=True))
        return 'OK'



    ####################################################################################################
    #### Log a problem with the SDRF file
    def log_problem(self, status, code, line, column, message):

        category = 'UNKNOWN'
        if status == 'WARNING':
            category = 'warnings'
        elif status == 'ERROR':
            category = 'errors'
        else:
            eprint(f"FATAL ERROR: Unrecognized event status '{status}'")
            eprint(sys.exc_info())
            sys.exit()

        #### Record the event
        full_message = f"{status}: [{code}]: {message} at line {line}, column {column}"
        self.problems[category]['count'] += 1
        self.problems[category]['list'].append(full_message)
        if code not in self.problems[category]['codes']:
            self.problems[category]['codes'][code] = 1
        else:
            self.problems[category]['codes'][code] += 1

        if self.verbose >= 1:
            eprint(full_message)



    ####################################################################################################
    #### Write the SDRF file
    def write(self, filename):

        if self.sdrf_data is None:
            eprint(f"ERROR: SDRFData.write: No data matrix content to write to '{filename}'")
            return

        if self.verbose >= 1:
            eprint(f"INFO: Writing SDRF file '{filename}'")

        with open(filename, 'w') as outfile:
            title_list = [str(value) for value in self.sdrf_data['titles']]
            print("\t".join(title_list), file=outfile)
            for row in self.sdrf_data['rows']:
                value_list = [str(value) for value in row]
                print("\t".join(value_list), file=outfile)



####################################################################################################
#### Main function for command-line usage
def main():

    import requests
    import requests_cache
    requests_cache.install_cache('sdrf_files_requests_cache.sqlite')

    #### Parse command line arguments
    argparser = argparse.ArgumentParser(description='Handler for SDRF files')
    argparser.add_argument('--url', action='store', help='URL of an SDRF file to download and read/validate')
    argparser.add_argument('--filepath', action='store', help='File path of an SDRF file to read/validate')
    argparser.add_argument('--merge_url', action='store', help='URL of an SDRF file to download, read/validate, and merge into the main SDRF (from --url or --filepath)')
    argparser.add_argument('--merge_filepath', action='store', help='File path of an SDRF file to read/validate and merge into the main SDRF (from --url or --filepath)')
    argparser.add_argument('--output_filepath', action='store', help='File path to which to write the resulting SDRF')
    argparser.add_argument('--tag_with_decisions', action='count', help='If set during a merge operation, all values are appended with a decision tag to indicate how that value came about from the decision making. This would not be used in production, but can be useful to debugging or assessing how to set parameters.')
    argparser.add_argument('--show_result', action='count', help='Show the JSON output of the reading or merging process (with the actual table dumped to an auxiliary TSV file)')
    argparser.add_argument('--prefer_main_columns', action='store', help='Comma separated list of columns for which the main file values should be preferred over the merge values')
    argparser.add_argument('--verbose', action='count' )
    params = argparser.parse_args()

    #### Create metadata handler object and read or create the metadata structure
    sdrf = SDRFHandler(verbose=params.verbose)
    sdrf_cache_path = "sdrf_files"

    if params.url is None and params.filepath is None:
        eprint(f"ERROR: Either --url or --filepath must be provided. Use --help for more info")
        return

    if params.url is not None:
        sdrf.data_url = params.url
        sdrf.ui_url = params.url

        response = requests.get(params.url)
        if response.status_code != 200:
            eprint(f"ERROR: Failed to download {params.url}. Status code: {response.status_code}")
            return
        if not os.path.exists(sdrf_cache_path):
            os.mkdir(sdrf_cache_path)

        filename = sdrf.extract_filename(params.url)
        filepath = f"{sdrf_cache_path}/{filename}"
        with open(filepath, 'wb') as outfile:
            outfile.write(response.content)

    if params.filepath is not None:
        filepath = params.filepath
        if not os.path.exists(filepath):
            eprint(f"ERROR: Specified filepath '{filepath}' does not exist")
            return

    #### Read the input SDRF and perform some light validation on it
    sdrf.read(filepath)

    #### If there are no merge operations requested, then just show results if requested and return
    if params.merge_url is None and params.merge_filepath is None:
        if params.show_result:
            tmp = sdrf.__dict__.copy()
            tmp['sdrf_data'] = "SDRF data structure truncated for display simplicity. See sdrf_handler_show_result.sdrf.tsv for contents"
            del(tmp['verbose'])
            print(json.dumps(tmp, indent=2, sort_keys=True))
            sdrf.write('sdrf_handler_show_result.sdrf.tsv')
        return


    #### Create a second SDRF object for the file to merge into the first (main) one
    merge_sdrf = SDRFHandler(verbose=params.verbose)

    #### If there's a merge_url specified, fetch it and store it locally
    if params.merge_url is not None:
        merge_sdrf.data_url = params.merge_url
        merge_sdrf.ui_url = params.merge_url

        response = requests.get(params.merge_url)
        if response.status_code != 200:
            eprint(f"ERROR: Failed to download {params.merge_url}. Status code: {response.status_code}")
            return

        merge_filename = sdrf.extract_filename(params.merge_url)
        merge_filepath = f"{sdrf_cache_path}/{merge_filename}"
        if merge_filepath == filepath:
            merge_filepath += '_to_merge'
        with open(merge_filepath, 'wb') as outfile:
            outfile.write(response.content)

    #### If there a merge filepath specified, ensure that it exists
    if params.merge_filepath is not None:
        merge_filepath = params.merge_filepath
        if not os.path.exists(merge_filepath):
            eprint(f"ERROR: Specified merge_filepath '{merge_filepath}' does not exist")
            return

    #### Read the second SDRF that will be merged into the first
    merge_sdrf.read(merge_filepath)

    #### Perform the merge of the merge_sdrf into the main one, passing user params to guide merging
    sdrf.merge(merge_sdrf, tag_with_decisions=params.tag_with_decisions, prefer_main_columns=params.prefer_main_columns)

    #### If requested, write out the result
    if params.output_filepath:
        sdrf.write(params.output_filepath)

    #### If the user requested to see the result of reading and merging
    if params.show_result:
        tmp = sdrf.__dict__.copy()
        tmp['sdrf_data'] = "SDRF data structure truncated for display simplicity. See sdrf_handler_show_result.sdrf.tsv for contents"
        del(tmp['verbose'])
        print(json.dumps(tmp, indent=2, sort_keys=True))
        sdrf.write('sdrf_handler_show_result.sdrf.tsv')



if __name__ == "__main__": main()



#!/usr/bin/env python3

import sys
import os
import argparse
import os.path
import timeit
import re
import gzip
def eprint(*args, **kwargs): print(*args, file=sys.stderr, **kwargs)

from pyteomics import mzml, auxiliary
from psims.transform.mzml import MzMLTransformer


def fmt(obj, formatstr):
    if obj is None:
        return ''
    else:
        return format(obj, formatstr)


####################################################################################################
class CustomMzMLTransformer(MzMLTransformer):
    def __init__(self, input_stream, output_stream, transform=None, transform_description=None,
                 sort_by_scan_time=False, scan_information=None, log_file_handle=None):
        super().__init__(input_stream, output_stream, transform=transform, transform_description=transform_description,
                 sort_by_scan_time=sort_by_scan_time)

        self.scan_information = scan_information
        self.log_file_handle = log_file_handle

    def format_spectrum(self, spectrum):
        new_spectrum = super().format_spectrum(spectrum)
        ms_level = None
        for param in new_spectrum['params']:
            if param['name'] == 'MS:1000511':
                ms_level = param['value']
        if ms_level > 1:
            #print(new_spectrum)
            nativeId = spectrum['id']
            scan_number = -1
            match = re.search(r'scan=(\d+)',nativeId)
            if match:
                scan_number = int(match.group(1))
            else:
                eprint(f"ERROR: Unable to extract scan number from {nativeId}")

            #print(f"scan={scan_number}")
            previous_precursor_mz = new_spectrum['precursor_information'][0]['mz']
            previous_charge = new_spectrum['precursor_information'][0]['charge']
            if scan_number in self.scan_information:
                new_precursor = self.scan_information[scan_number]['precursor_mz']
                new_charge = self.scan_information[scan_number]['charge_state']

                new_spectrum['precursor_information'][0]['mz'] = new_precursor
                new_spectrum['precursor_information'][0]['charge'] = new_charge
            else:
                print(f"WARNING: Did not have recorded updated information for scan {scan_number}")

            #### This fails if the previous_charge is None
            #print(f"{scan_number:6d}\t{previous_precursor_mz:10.4f}\t{previous_charge:2d}\t{new_precursor:10.4f}\t" +
            #    f"{new_charge:2d}\t{float(new_precursor)-float(previous_precursor_mz):10.4f}\t{int(new_charge)-int(previous_charge):2d}", file=self.log_file_handle )

            #### Special handling for a charge that is null
            fixed_previous_charge = previous_charge
            if previous_charge is None:
                fixed_previous_charge = 0

            #### Write details of the before and after to the log file
            print(f"{fmt(scan_number,'6d')}\t{fmt(previous_precursor_mz,'10.4f')}\t{fmt(previous_charge,'2d')}\t{fmt(new_precursor,'10.4f')}\t" +
                f"{fmt(new_charge,'2d')}\t{fmt(float(new_precursor)-float(previous_precursor_mz),'10.4f')}\t{fmt(int(new_charge)-int(fixed_previous_charge),'2d')}", file=self.log_file_handle )

        return new_spectrum




####################################################################################################
#### Monocle CSV Reader class reads a Monocle CSV and extracts the precursor m/z values and charge values
class MonocleCsvReader:


    ####################################################################################################
    #### Constructor
    def __init__(self, csv_file, verbose=None):

        self.csv_file = csv_file

        #### General stats
        self.stats = {}

        #### Information about each scan
        self.scans = {}

        #### Set verbosity
        if verbose is None: verbose = 0
        self.verbose = verbose


    ####################################################################################################
    #### Read spectra
    def read_spectra(self):

        #### Set up information
        t0 = timeit.default_timer()
        stats = {
            'n_spectra': 0,
            'n_ms0_spectra': 0,
            'n_ms1_spectra': 0,
            'n_ms2_spectra': 0,
            'n_ms3_spectra': 0,
            'n_length_zero_spectra': 0
        }
        self.stats = stats

        #### Show information
        if self.verbose >= 1:
            eprint(f"INFO: Reading Monocle CSV file {self.csv_file}")

        #### Read spectra from the file
        with open(self.csv_file) as infile:
            iline = 0
            for line in infile:
                iline += 1
                if iline == 1:
                    if not line.startswith('scan number,ms level'):
                        print(f"ERROR: CSV file header line not as expected: {line}")
                        exit()
                    else:
                        continue

                columns = line.strip().split(',')

                #### Extract the MS level of the spectrum
                scan_number = int(columns[0])
                ms_level = int(columns[1])
                ms_level_stat = f"n_ms{ms_level}_spectra"
                stats[ms_level_stat] += 1

                precursor_mz = float(columns[2])
                charge_state = int(columns[4])

                charge_stat = f"n_charge_{charge_state}_precursors"
                if charge_stat not in stats:
                    stats[charge_stat] = 0
                stats[charge_stat] += 1

                #### Store the supplied precursor_ms and charge state in the cache
                self.scans[scan_number] = {'precursor_mz': precursor_mz, 'charge_state': charge_state }

                #### Update counters
                stats['n_spectra'] += 1

        infile.close()

        #### Print final timing information
        t1 = timeit.default_timer()
        print(f"INFO: Read {stats['n_spectra']} spectra from {self.csv_file} in {int(t1-t0)} sec ({stats['n_spectra']/(t1-t0)} spectra per sec)")




####################################################################################################
#### Monocle mzML Reader class reads an mzML and extracts the precursor m/z values and charge values
class MonocleMzmlReader:


    ####################################################################################################
    #### Constructor
    def __init__(self, mzml_file, verbose=None):

        self.mzml_file = mzml_file

        #### General stats
        self.stats = {}

        #### Information about each scan
        self.scans = {}

        #### Set verbosity
        if verbose is None: verbose = 0
        self.verbose = verbose


    ####################################################################################################
    #### Read spectra
    def read_spectra(self):

        #### Set up information
        t0 = timeit.default_timer()
        stats = {
            'n_spectra': 0,
            'n_ms0_spectra': 0,
            'n_ms1_spectra': 0,
            'n_ms2_spectra': 0,
            'n_ms3_spectra': 0,
            'n_length_zero_spectra': 0
        }
        self.stats = stats

        #### Show information
        if self.verbose >= 1:
            eprint(f"INFO: Reading Monocle mzML file {self.mzml_file}", end='', flush=True)
            progress_intro = False

        #### If the mzML is gzipped, then open with zlib, else a plain open
        if self.mzml_file.endswith('.gz'):
            infile = gzip.open(self.mzml_file)
        else:
            infile = open(self.mzml_file, 'rb')

        #### Read spectra from the file
        with mzml.read(infile) as reader:
            #try:
            if True:
                for spectrum in reader:

                    #### Debugging. Print the data structure of the first spectrum
                    #if stats['n_spectra'] == 0:
                    #    auxiliary.print_tree(spectrum)
                    #    print(spectrum)
                    #    return

                    #### Extract the MS level of the spectrum
                    ms_level = spectrum['ms level']
                    ms_level_stat = f"n_ms{ms_level}_spectra"
                    stats[ms_level_stat] += 1

                    #### Save the spectrum in our cache
                    nativeId = spectrum['id']
                    scan_number = -1
                    match = re.search(r'scan=(\d+)',nativeId)
                    if match:
                        scan_number = int(match.group(1))
                    else:
                        eprint(f"ERROR: Unable to extract scan number from {nativeId}")

                    #### If the ms level is 2, then examine it for information
                    if ms_level == 2 and 'm/z array' in spectrum:
                        #### Monocle outputs the wrong terms
                        #precursor_mz = spectrum['precursorList']['precursor'][0]['selectedIonList']['selectedIon'][0]['selected ion m/z']
                        #charge_state = int(spectrum['precursorList']['precursor'][0]['selectedIonList']['selectedIon'][0]['charge state'])
                        precursor_mz = spectrum['precursorList']['precursor'][0]['selectedIonList']['selectedIon'][0]['isolation window target m/z']
                        charge_state = int(spectrum['precursorList']['precursor'][0]['selectedIonList']['selectedIon'][0]['possible charge state'])

                        charge_stat = f"n_charge_{charge_state}_precursors"
                        if charge_stat not in stats:
                            stats[charge_stat] = 0
                        stats[charge_stat] += 1

                        #### Store the supplied precursor_ms and charge state in the cache
                        self.scans[scan_number] = {'precursor_mz': precursor_mz, 'charge_state': charge_state }

                        #### Check for zero length and very sparse spectra
                        if len(spectrum['m/z array']) == 0:
                            stats['n_length_zero_spectra'] += 1

                    #### Update counters and print progress
                    stats['n_spectra'] += 1
                    if self.verbose >= 1:
                        if stats['n_spectra']/500 == int(stats['n_spectra']/500):
                            if not progress_intro:
                                #eprint("INFO: Reading spectra.. ", end='')
                                progress_intro = True
                            #eprint(f"{stats['n_spectra']}.. ", end='', flush=True)
                            eprint(".", end='', flush=True)
            #except:
            else:
                self.log_event('ERROR','MzMLCorrupt',f"Pyteomics threw an error reading mzML file! File may be corrupt. Check file '{self.mzml_file}'")

        infile.close()
        if self.verbose >= 1:
            eprint("")

        #### Print final timing information
        t1 = timeit.default_timer()
        print(f"INFO: Read {stats['n_spectra']} spectra from {self.mzml_file} in {int(t1-t0)} sec ({stats['n_spectra']/(t1-t0)} spectra per sec)")


####################################################################################################
#### For command-line usage
def main():

    argparser = argparse.ArgumentParser(description='Read and write an mzML file')
    argparser.add_argument('--verbose', action='count', help='If set, print more information about ongoing processing' )
    argparser.add_argument('--input_filename', type=str, action='store', required=True, help='Name of the input ThermoRawFileParser mzML file')
    argparser.add_argument('--monocle_filename', type=str, action='store', required=True, help='Name of the Monocle csv or mzML output from which to get precursor information')
    argparser.add_argument('--output_filename', type=str, action='store', required=True, help='Name of the output mzML based on the TRFP mzML but with Monocle precursor information')
    params = argparser.parse_args()

    #### Set verbose
    verbose = params.verbose
    if verbose is None:
        verbose = 1

    #### Check that the filenames are defined and exist
    if not os.path.isfile(params.input_filename):
        print(f"ERROR: --input_filename '{params.input_filename}' not found or not a file")
        return
    if not os.path.isfile(params.monocle_filename):
        print(f"ERROR: --monocle_filename '{params.monocle_filename}' not found or not a file")
        return

    #### Ensure the output is not the same as the input
    if params.output_filename == params.input_filename:
        print(f"ERROR: --output_filename '{params.input_filename}' may not be the same as the --input_filename")
        return
    if params.output_filename == params.monocle_filename:
        print(f"ERROR: --monocle_filename '{params.monocle_filename}' may not be the same as the --input_filename")
        return

    #### If the Monocle file is an mzML, read and extract information from that
    if params.monocle_filename.endswith('.mzML') or params.monocle_filename.endswith('.mzML.gz'):
        reader = MonocleMzmlReader(params.monocle_filename, verbose=verbose)
        reader.read_spectra()
    elif params.monocle_filename.endswith('.csv'):
        reader = MonocleCsvReader(params.monocle_filename, verbose=verbose)
        reader.read_spectra()
    else:
        print(f"ERROR: Expected Monocle file to have extensions .csv or .mzML or .mzML.gz but '{params.monocle_filename}' does not")
        return

    #### Open in the input file
    print(f"INFO: Opening mzML file '{params.input_filename}' for reading")
    if params.input_filename.endswith('.gz'):
        infile = gzip.open(params.input_filename)
    else:
        infile = open(params.input_filename, 'rb')

    #### Open the output file
    try:
        outfile = open(params.output_filename, 'wb')
    except:
        print(f"ERROR: --output_filename '{params.output_filename}' is not writable. Permissions problem?")
        return

    log_file_handle = open(params.output_filename+'.log', 'w')
    print("scan_number\tprevious_precursor_mz\tprevious_charge\tnew_precursor_mz\t" +
          "new_charge\tprecursor_mz_delta\tcharge_delta", file=log_file_handle )

    print(f"INFO: Opening {params.output_filename} for writing")
    CustomMzMLTransformer(infile, outfile, scan_information=reader.scans, log_file_handle=log_file_handle,
        transform_description='Transfer Monocle precursor refinements to well-formed mzML').write()


#### For command line usage
if __name__ == "__main__": main()



#!/usr/bin/env python3

import sys
import os
import argparse
import multiprocessing
import copy
from datetime import datetime
import timeit
from matplotlib.backends.backend_pdf import PdfPages
from pypdf import PdfReader, PdfWriter

from runassessor.mzML_assessor import MzMLAssessor
from runassessor.metadata_handler import MetadataHandler
from runassessor.graph_generator import GraphGenerator

def eprint(*args, **kwargs): print(*args, file=sys.stderr, **kwargs)


####################################################################################################
#### Process one mzML
def process_job(job):

    #### Assess the mzML file
    assessor = MzMLAssessor(job['filename'], metadata=job['metadata'], verbose=job['verbose'], frag_units=job['fragmentation_units'])
    assessor.read_header()
    assessor.read_spectra(write_fragmentation_type_file=job['write_fragmentation_type_files'])
    assessor.assess_lowend_composite_spectra()
    assessor.assess_neutral_loss_composite_spectra()

    if assessor.metadata['state']['status'] != 'ERROR' or 'multiple fragmentation types' in assessor.metadata['state']['message']:
        assessor.assess_ROIs()
        
        
    return assessor



####################################################################################################
#### Main function for command-line usage
def main():

    argparser = argparse.ArgumentParser(description='Read one or more mzML files and extract some knowledge from them')
    argparser.add_argument('--metadata_filepath', action='store', help='Filename and path of the ouput metadata file (defaults to study_metadata.json in the current working directory)')
    argparser.add_argument('--preserve_existing_metadata', action='count', default=0, help='If set, read the existing metadata file, preserve its contents, and build on it (default is to overwrite current file)')
    argparser.add_argument('--n_threads', action='store', type=int, help='Set the number of files to process in parallel (defaults to number of cores)')
    argparser.add_argument('--write_fragmentation_type_files', action='count', help='If set, write a fragmentation_type file for each mzML')
    argparser.add_argument('--verbose', action='count', help='If set, print more information about ongoing processing' )
    argparser.add_argument('--version', action='version', version='%(prog)s 0.8')
    argparser.add_argument('--write_sdrf_file', action='count', help='If set, then write an SDRF file based on the metadata file')
    argparser.add_argument('--include_sdrf_provenance', action='count', help='If set, then all information written to the SDRF file will have provenance tags (non-standard SDRF)')
    argparser.add_argument('--write_pdfs', action='count', help='If set, then generate a PDF of delta time and ppm graphs from precursor stats')
    argparser.add_argument('--write_ions', action='count', help='If set, then write a TSV file of fragment ion peaks found')
    argparser.add_argument('--ms_ms_units', choices=['mz', 'ppm'], help='Sets units for MS/MS tolerance recommendations: mz or ppm (defaults to ppm for HR and mz for LR)')
    argparser.add_argument('files', type=str, nargs='+', help='Filenames of one or more mzML files to read')
    params = argparser.parse_args()

    #### Set verbose level
    verbose = params.verbose
    if verbose is None:
        verbose = 0
    else:
        verbose = 1
    if verbose >= 1:
        timestamp = str(datetime.now().isoformat())
        eprint(f"INFO: Launching RunAssessor at {timestamp}")
    t0 = timeit.default_timer()

    #### Loop over all the files to ensure that they are really there before starting work
    n_files = 0
    for file in params.files:
        if not os.path.isfile(file):
            eprint(f"ERROR: File '{file}' not found or not a file")
            return
        n_files += 1
    if verbose:
        eprint(f"INFO: Found {n_files} input files to process")

    #### Initialize the metadata handler
    study = MetadataHandler(params.metadata_filepath, verbose=params.verbose, write_ions=params.write_ions, frag_units=params.ms_ms_units)

    #### If the user wants to read and preserve an existing file, then read or create it
    if params.preserve_existing_metadata > 0:
        result = study.read_or_create()
    else:
        result = study.create()
    if result is None or result != 'OK':
        eprint(f"ERROR: Unable to initialize study metadata information. Halting.")
        return

    #### Set up the list of jobs to process
    jobs = []

    #### Loop over all the files and put them in a list of jobs for parallel processing
    for file in params.files:

        #### If this file is already in the metadata store, purge it and generate new
        if file in study.metadata['files']:
            if params.preserve_existing_metadata > 0:
                if verbose >= 1:
                    eprint(f"INFO: Already have results for '{file}'. Skipping..")
                continue
            else:
                study.metadata['files'][file] = None

        job = {
            'filename': file,
            'metadata': copy.deepcopy(study.metadata),
            'verbose': verbose,
            'write_fragmentation_type_files': params.write_fragmentation_type_files,
            'fragmentation_units': params.ms_ms_units
            }
        jobs.append(job)


    #### Compute how many files in parallel to process
    n_cpus = multiprocessing.cpu_count()
    n_simultaneous_jobs = n_cpus
    if n_files < n_simultaneous_jobs:
        n_simultaneous_jobs = n_files
    n_threads = params.n_threads
    if n_threads is None or n_threads == 0:
        n_threads = n_simultaneous_jobs
    elif n_threads < 0:
        if -1 * n_threads > n_cpus - 1:
            eprint(f"ERROR: Parameter n_threads ({n_threads} too low for number of CPUs ({n_cpus}))")
            return
        n_threads = n_cpus + n_threads
        if n_threads < n_simultaneous_jobs:
            n_threads = n_files


    #### Now process the jobs in parallel
    if n_threads == 1:
        eprint(f"INFO: Processing 1 input file", end='', flush=True)
    else:
        eprint(f"INFO: Processing {n_threads} files in parallel with multiprocessing (one file per CPU)", end='', flush=True)
    pool = multiprocessing.Pool(processes=n_threads)
    results = pool.map_async(process_job, jobs)
    pool.close()
    pool.join()
    eprint('')

    #### Coalesce the results
    results = results.get()
    for result in results:
        for filename in result.metadata['files']:
            study.metadata['files'][filename] = result.metadata['files'][filename]
        study.metadata['state'] = result.metadata['state']
        study.metadata['state']['problems'] = result.metadata['problems']

    #### Infer parameters based on the latest data
    study.infer_search_criteria()
    study.generate_sdrf_table(include_provenance=params.include_sdrf_provenance)

    #### Write out our state of mind
    study.store()

    #### Write out SDRF table if parameter given
    if params.write_sdrf_file:
        key_value_file = study.find_txt_file()
        study.read_txt_file(key_value_file)
        sdrf_filename = study.infer_sdrf_filename()
        study.write_sdrf_file(sdrf_filename) 

    if verbose >= 1:
        timestamp = str(datetime.now().isoformat())
        t1 = timeit.default_timer()
        eprint(f"INFO: RunAssessor finished in {t1-t0:.2f} seconds at {timestamp}")
    
    #### Write out graphs is parameter given
    if params.write_pdfs:
        grapher = GraphGenerator(params.metadata_filepath, verbose=params.verbose)
        metadata_name = grapher.buildGraphs()
        eprint("INFO: Delta Graphs generated and stored")

        nl_pdf = metadata_name.replace(".json", ".NLplots.pdf")
        # Write cover page documentation to pdf
        nl_coverpage = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "lib", "neutral_loss_windows_documentation.pdf")
        writer = PdfWriter()
        cover_reader = PdfReader(nl_coverpage)
        for page in cover_reader.pages:
            writer.add_page(page)
        # Generate neutral loss windows and add to the pdf
        with PdfPages(nl_pdf) as pdf:
            for assessor in results:
                grapher.plot_precursor_loss_composite_spectra(assessor=assessor, pdf=pdf)
        output_reader = PdfReader(nl_pdf)
        for page in output_reader.pages:
            writer.add_page(page)
        with open(nl_pdf, "wb") as f:
            writer.write(f)
        eprint("INFO: Neutral Loss Graphs generated and stored")


if __name__ == "__main__": main()



import json
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import norm
from math import sqrt, pi
from matplotlib.backends.backend_pdf import PdfPages
from pypdf import PdfReader, PdfWriter
import sys
import os
def eprint(*args, **kwargs): print(*args, file=sys.stderr, **kwargs)

class GraphGenerator:
    def __init__(self, metadata_filepath=None, verbose=None):
        default_metadata_filepath = 'study_metadata.json'
        if metadata_filepath == None:
            self.metadata_file = default_metadata_filepath
        elif not metadata_filepath.endswith('.json'):
            self.metadata_file = metadata_filepath + ".json"
        else:
            self.metadata_file = metadata_filepath

         #### Load JSON data
        try:
            with open(self.metadata_file, "r") as f:
                self.data = json.load(f)
        except:
            eprint("INFO: Unable to read JSON file")
            return
        
        if verbose != None and verbose != 0:
            self.verbose = 1
        else:
            self.verbose = 0
        
        self.files = {}

    def fmt(self, v):
        return f"{v:.3f}" if isinstance(v, (int, float)) else str(v)

    def buildGraphs(self):

        self.files = self.data.get("files", {})

        coverpage = os.path.join(os.path.dirname(os.path.abspath(__file__)), "Delta_graphs_documentation.pdf")

        #### Output PDF file
        output_pdf_path = self.metadata_file.replace(".json", ".precursor_histograms.pdf")
        with PdfPages(output_pdf_path) as pdf:

            for filename, file_data in self.files.items():
                if file_data.get("spectra_stats", {}).get("acquisition_type", 'none') == "DIA":
                    continue
                pre_tol = {}
                hist_ppm = {}
                amplitude_ppm = None
                mu_ppm = None
                sigma_ppm = None
                y_offset_ppm = None
                chi_squared_ppm = None

                hist_time = {}
                time_initial_level = None
                pulse_start = None
                peak_level = None
                decay_constant = None
                final_level = None
                #pulse_duration = None
                chi_squared_time = None

                time_bin_centers = None
                time_counts = None
                ppm_bin_centers = None
                ppm_counts = None
                precursor_stats = file_data.get("summary", {}).get('precursor stats', {})
                dynamic_exclusion_window = precursor_stats.get('dynamic exclusion window', {})

                try:
                    pre_tol = precursor_stats.get('precursor tolerance', {})
                    hist_ppm = pre_tol.get("histogram_ppm", {})
                    amplitude_ppm = pre_tol.get('fit_ppm', {}).get("intensity", [])
                    mu_ppm = pre_tol.get('fit_ppm', {}).get("delta_ppm peak", [])
                    sigma_ppm = pre_tol.get('fit_ppm', {}).get("sigma (ppm)", [])
                    y_offset_ppm = pre_tol.get('fit_ppm', {}).get("y_offset", [])
                    
                    chi_squared_ppm = pre_tol.get("fit_ppm", {}).get("chi_squared", "N/A")
                except:
                    pass

                try:
                    hist_time = dynamic_exclusion_window.get("histogram_time", {})
                    time_initial_level = dynamic_exclusion_window.get("fit_pulse_time", {}).get("inital level", [])
                    pulse_start = dynamic_exclusion_window.get("fit_pulse_time", {}).get("pulse start", [])
                    peak_level = dynamic_exclusion_window.get("fit_pulse_time", {}).get("peak level", [])
                    decay_constant = dynamic_exclusion_window.get("fit_pulse_time", {}).get("decay constant", [])
                    final_level = dynamic_exclusion_window.get("fit_pulse_time", {}).get("final level", [])
                    #duration = dynamic_exclusion_window.get("fit_pulse_time", {}).get("pulse duration", [])
                    chi_squared_time = dynamic_exclusion_window.get("fit_pulse_time", {}).get("chi_squared", "N/A")
                except:
                    pass

                

                try:
                    time_bin_centers = np.array(hist_time.get("bin_centers", []))
                    time_counts = np.array(hist_time.get("counts", []))
                except:
                    pass
                
                try:
                    ppm_bin_centers = np.array(hist_ppm.get("bin_centers", []))
                    ppm_counts = np.array(hist_ppm.get("counts", []))
                except:
                    pass


                fig, axs = plt.subplots(2, 2, figsize=(12, 8), gridspec_kw={'height_ratios': [4, 1]})
                fig.suptitle(filename, fontsize=14)

                #### Time histogram
                ax_time = axs[0, 0]
                try:
                    width_time = (time_bin_centers[1] - time_bin_centers[0]) if len(time_bin_centers) > 1 else 1
                    ax_time.bar(time_bin_centers, time_counts, width=width_time, color='skyblue', edgecolor='black', label='Time Data')
                
                    try:
                        if peak_level <= 0:
                            pass
                        else:
                            t_fine = np.linspace(min(time_bin_centers), max(time_bin_centers), 500)
                            y = np.full_like(t_fine, time_initial_level, dtype=float)

                            pulse_end = pulse_start
                            pulse_mask = (t_fine >= pulse_start) & (t_fine < pulse_end)
                            y[pulse_mask] = peak_level
                            decay_mask = t_fine >= pulse_end
                            safe_tau = max(decay_constant, 1e-8)
                            exponent = -(t_fine[decay_mask] - pulse_end) / safe_tau
                            exponent = np.clip(exponent, -700, 700)
                            y[decay_mask] = final_level + (peak_level - final_level) * np.exp(exponent)

                            ax_time.plot(t_fine, y, color='blue', linewidth=2, label='Time Decay Fit')
                    except:
                        pass
                except:
                    ax_time.bar([], [])

                ax_time.set_title("ΔTime Between Precursors")
                ax_time.set_xlabel("ΔTime (seconds)")
                ax_time.set_ylabel("Counts")
                ax_time.legend()

                #### PPM histogram
                ax_ppm = axs[0, 1]
                try: 
                    width_ppm = (ppm_bin_centers[1] - ppm_bin_centers[0]) if len(ppm_bin_centers) > 1 else 1
                    ax_ppm.bar(ppm_bin_centers, ppm_counts, width=width_ppm, color='salmon', edgecolor='black', label='PPM Data')

                
                    try:
                        
                        if amplitude_ppm <= 0:
                            pass
                        else:
                            x_ppm = np.linspace(min(ppm_bin_centers), max(ppm_bin_centers), 500)
                            scale = amplitude_ppm * sigma_ppm * sqrt(2 * pi)
                            y_ppm = scale * norm.pdf(x_ppm, mu_ppm, sigma_ppm) + y_offset_ppm
                            ax_ppm.plot(x_ppm, y_ppm, color='darkred', linewidth=2, label='PPM Gaussian Fit')
                    except:
                        pass
                except:
                    ax_ppm.bar([], [])
                    
                ax_ppm.set_title("Precursor m/z Variance in PPM")
                ax_ppm.set_xlabel("PPM Variance")
                ax_ppm.set_ylabel("Counts")
                ax_ppm.legend()

                #### Tables
                axs[1, 0].axis('off')
                axs[1, 1].axis('off')


                ## Table under Time histogram
                try:
                    if peak_level == None or peak_level == 0:
                        time_table_data = [["No fit found", ""]]
                    else:
                        time_table_data = [
                            ["Chi² Time", self.fmt(chi_squared_time)],
                            ["Initial Level", self.fmt(time_initial_level)],
                            ["Pulse Start", self.fmt(pulse_start)],
                            ["Peak Level",self.fmt(peak_level)],
                            ["Decay Constant",self.fmt(decay_constant)],
                            ["Final Level", self.fmt(final_level)],
                        ]
                except:
                    time_table_data = [["No fit found", ""]]
                    
                table_time = axs[1, 0].table(
                        cellText=time_table_data,
                        colLabels=["Time Fit", "Value"],
                        loc='center',
                        cellLoc='left',
                        colWidths=[0.35, 0.3],
                        bbox=[0, -0.25, 1, 1] 
                    )
                table_time.scale(1.2, 3)

                ## Table under PPM histogram
                try:
                    if amplitude_ppm == None or amplitude_ppm  == 0:
                        ppm_table_data = [["No fit found", ""]]
                    else:
                        ppm_table_data = [
                            ["Chi² PPM", self.fmt(chi_squared_ppm)],
                            ["Amplitude", self.fmt(amplitude_ppm)],
                            ["Mu (Peak Center)", self.fmt(mu_ppm)],
                            ["Sigma", self.fmt(sigma_ppm)],
                            ["Y Offset", self.fmt(y_offset_ppm)],
                        ]
                except:
                    ppm_table_data = [["No fit found", ""]]

                table_ppm = axs[1, 1].table(
                    cellText=ppm_table_data,
                    colLabels=["PPM Fit", "Value"],
                    loc='center',
                    cellLoc='left',
                    colWidths=[0.35, 0.3],
                    bbox=[0, -0.25, 1, 1]  
                )
                table_ppm.scale(1.2, 1.6)
               
                    


        
                pdf.savefig(fig)
                plt.close(fig)

            if self.verbose >= 1:
                print(f"Saved plot for {filename}")

        #Add explanatory Cover Page
        writer = PdfWriter()
        cover_reader = PdfReader(coverpage)
        for page in cover_reader.pages:
            writer.add_page(page)

        try:
            output_reader = PdfReader(output_pdf_path)
            for page in output_reader.pages:
                writer.add_page(page)

            # Write the final combined PDF
            with open(output_pdf_path, "wb") as f:
                writer.write(f)
        except:
            eprint("WARNING: No doumentation page added to chi_sqared graphs")

        if self.verbose >= 1:
            eprint("Saving PDF to:", os.path.abspath(output_pdf_path))
        return self.metadata_file


    
    def plot_precursor_loss_composite_spectra(self, assessor, pdf):
        destinations_list = list(assessor.composite.keys())
        file_name_root = assessor.mzml_file.split('.')[0]
        for destination in destinations_list:
            if destination.startswith("precursor_loss_"):

                intensities = assessor.composite[destination]['intensities']
                maximum = assessor.composite[destination]['maximum']
                minimum = assessor.composite[destination]['minimum']
                binsize = assessor.composite[destination]['binsize']
                mz = np.arange((maximum - minimum) / binsize + 1) * binsize + minimum

                water_z2 = 9.00528235
                phosphoric_acid_z2 = 48.98844785
                try:
                    water_z2_extent_bins = assessor.metadata['files'][assessor.mzml_file]['neutral_loss_peaks'][destination]['water_z2']['peak']['extended']['extent']
                except:
                    water_z2_extent_bins = 0
                try:
                    phosphoric_acid_z2_extent_bins = assessor.metadata['files'][assessor.mzml_file]['neutral_loss_peaks'][destination]['phosphoric_acid_z2']['peak']['extended']['extent']
                except:
                    phosphoric_acid_z2_extent_bins = 0

                water_z2_axis_min = water_z2 - 30 * binsize
                water_z2_axis_max = water_z2 + 30 * binsize
                water_z2_extent_min = water_z2 - water_z2_extent_bins * binsize
                water_z2_extent_max = water_z2 + water_z2_extent_bins * binsize
                water_z2_fitting_min = water_z2 - 2 * water_z2_extent_bins * binsize
                water_z2_fitting_max = water_z2 + 2 * water_z2_extent_bins * binsize

                phosphoric_acid_z2_axis_min = phosphoric_acid_z2 - 30 * binsize
                phosphoric_acid_z2_axis_max = phosphoric_acid_z2 + 30 * binsize
                phosphoric_acid_z2_extent_min = phosphoric_acid_z2 - phosphoric_acid_z2_extent_bins * binsize
                phosphoric_acid_z2_extent_max = phosphoric_acid_z2 + phosphoric_acid_z2_extent_bins * binsize
                phosphoric_acid_z2_fitting_min = phosphoric_acid_z2 - 2 * phosphoric_acid_z2_extent_bins * binsize
                phosphoric_acid_z2_fitting_max = phosphoric_acid_z2 + 2 * phosphoric_acid_z2_extent_bins * binsize

                sum_type = destination.replace("precursor_loss_", "")
                summary = self.files[assessor.mzml_file]['summary'][sum_type]

                # Colors based on detection
                water_color = 'palegreen' if summary.get('has water_loss') else 'navajowhite'
                phospho_color = 'palegreen' if summary.get('has phospho_spectra') else 'navajowhite'

                correct_bar_water = 'limegreen' if summary.get('has water_loss') else 'orange'
                correct_bar_phospho = 'limegreen' if summary.get('has phospho_spectra') else 'orange'

                ratio = summary.get('z=2 phosphoric_acid to z=2 water_loss intensity ratio', 'N/A')
                num_water = assessor.metadata['files'][assessor.mzml_file]['neutral_loss_peaks'][destination]['water_z2']['peak']['mode_bin']['n_spectra']
                num_phospho = assessor.metadata['files'][assessor.mzml_file]['neutral_loss_peaks'][destination]['phosphoric_acid_z2']['peak']['mode_bin']['n_spectra']
                abs_diff_delta_mz = summary.get('absolute difference in delta m/z for z=2 phosphoric_acid_loss and z=2 water_loss', 'N/A')

                # Plot water z=2
                if water_z2_extent_bins * 2 < 30:
                    plt.axvspan(xmin=water_z2_fitting_min, xmax=water_z2_fitting_max, color='lightsteelblue', alpha=0.25, lw=0)
                plt.axvspan(xmin=water_z2_extent_min, xmax=water_z2_extent_max, color=water_color, alpha=0.75, lw=0)
                plt.axvline(x=water_z2, color=correct_bar_water, alpha=0.75)
                water_z2_range = (mz >= water_z2_axis_min) & (mz <= water_z2_axis_max)
                plt.plot(mz[water_z2_range], intensities[water_z2_range])
                
                ## Plot possible fit
                little_label = ''
                try:
                    if self.files[assessor.mzml_file]['neutral_loss_peaks'][destination]['water_z2']['peak']['assessment']['is_found']:
                        fit_param = self.files[assessor.mzml_file]['neutral_loss_peaks'][destination]['water_z2']['peak']['fit']
                        sigma_mz = fit_param['sigma_mz']
                        mu_mz = fit_param['mz']
                        y_offset = fit_param['y_offset']
                        mz_range = mz[water_z2_range]
                        mz_subset = np.linspace(min(mz_range), max(mz_range), 1000)
                        intensity_subset = intensities[water_z2_range]

                        gaussian = norm.pdf(mz_subset, mu_mz, sigma_mz)
                        scaled_gaussian = gaussian * np.max(intensity_subset) / np.max(gaussian) + y_offset/ np.max(gaussian)

                        plt.plot(mz_subset, scaled_gaussian, color='darkred', linewidth=2)
                    else:
                        if self.verbose >= 1:
                            eprint(f"No water loss z =2 peak found in {assessor.mzml_file}")
                        little_label = " (no fit found)"
                except:
                    pass
                plt.xlabel("m/z loss (water z=2)" + little_label + "\nMode of water z=2 spectra: " + str(num_water))
                plt.ylabel("Intensity")
                try:
                    plt.title(f"{file_name_root} ({destination})\nAbsolute difference between delta m/z phosphoric acid and water: {abs_diff_delta_mz:.4f}", fontsize=8.5)
                except:
                    plt.title(f"{file_name_root} ({destination})\nAbsolute difference between delta m/z phosphoric acid and water: {abs_diff_delta_mz}", fontsize=8.5)
                plt.tight_layout()
                pdf.savefig()
                plt.close()

                # Plot phosphoric acid z=2
                if phosphoric_acid_z2_extent_bins * 2 < 30:
                    plt.axvspan(xmin=phosphoric_acid_z2_fitting_min, xmax=phosphoric_acid_z2_fitting_max, color='lightsteelblue', alpha=0.25, lw=0)
                plt.axvspan(xmin=phosphoric_acid_z2_extent_min, xmax=phosphoric_acid_z2_extent_max, color=phospho_color, alpha=0.75, lw=0)
                plt.axvline(x=phosphoric_acid_z2, color=correct_bar_phospho, alpha=0.75)
                phosphoric_acid_z2_range = (mz >= phosphoric_acid_z2_axis_min) & (mz <= phosphoric_acid_z2_axis_max)
                plt.plot(mz[phosphoric_acid_z2_range], intensities[phosphoric_acid_z2_range])
                plt.xlabel("m/z loss (phosphoric acid z=2)")
                little_label = ""
                try:
                    if self.files[assessor.mzml_file]['neutral_loss_peaks'][destination]['phosphoric_acid_z2']['peak']['assessment']['is_found']:
                        fit_param = self.files[assessor.mzml_file]['neutral_loss_peaks'][destination]['phosphoric_acid_z2']['peak']['fit']
                        sigma_mz = fit_param['sigma_mz']
                        mu_mz = fit_param['mz']
                        y_offset = fit_param['y_offset']
                        mz_range = mz[phosphoric_acid_z2_range]
                        mz_subset = np.linspace(min(mz_range), max(mz_range), 1000)
                        intensity_subset = intensities[phosphoric_acid_z2_range]

                        gaussian = norm.pdf(mz_subset, mu_mz, sigma_mz)
                        scaled_gaussian = gaussian * np.max(intensity_subset) / np.max(gaussian) + y_offset

                        plt.plot(mz_subset, scaled_gaussian, color='darkred', linewidth=2)
                    else:
                        if self.verbose >= 1:
                            eprint(f"No phosphoric acid z =2 peak found in {assessor.mzml_file}")
                        little_label = " (no fit found)"
                except:
                    pass

                plt.xlabel("m/z loss (phosphoric acid z=2)" + little_label + "\nMode of phosphoric acid z=2 spectra: " + str(num_phospho))
                plt.ylabel("Intensity")
                try:
                    plt.title(f"{file_name_root} ({destination})\nAbsolute difference between delta m/z phosphoric acid and water: {abs_diff_delta_mz:.4f}\nPhosphoric acid to water intensity ratio: {ratio:.2f}", fontsize=8.5)
                except:
                    try:
                        plt.title(f"{file_name_root} ({destination})\nAbsolute difference between delta m/z phosphoric acid and water: {abs_diff_delta_mz}\nPhosphoric acid to water intensity ratio: {ratio:.2f}", fontsize=8.5)
                    except:
                        plt.title(f"{file_name_root} ({destination})\nAbsolute difference between delta m/z phosphoric acid and water: {abs_diff_delta_mz}\nPhosphoric acid to water intensity ratio: {ratio}", fontsize=8.5)
                plt.tight_layout()
                pdf.savefig()
                plt.close()

                # Full neutral loss spectrum
                plt.plot(mz, intensities)
                plt.xlabel("m/z loss")
                plt.ylabel("Intensity")
                plt.title(f"{file_name_root} ({destination})\nFull Neutral Loss Spectrum", fontsize=8.5)
                plt.tight_layout()
                pdf.savefig()
                plt.close()
        nl_pdf = self.metadata_file.replace(".json", ".NLplots.pdf")
        if self.verbose >= 1:
            eprint(f"All {assessor.mzml_file} neutral loss spectra saved to: {nl_pdf}")
